// Sketch to read up to 10 DS18B20 temperature sensors on the heat bank and regulate when the 
// backup gas boiler fires and for how long. 

// The sketch publishes the temperatures to the local emoncms server which makes all the data
// available outside the house (used for dashboards etc.). This is done every 15 seconds using a timer.

// The sketch includes a webserver. This is used to turn the CH on and off (works in 
// parallel with the time controller on the Heat Bank controls). The web pages also show temperatures for the setpoint and 
// the temperatures of the feeds to the UFH and rads. In addition there are pages to alter the boiler demand set point
// and to ensure the temperature sensor allocation is set up properly. 

// The sketch controls the backup boiler for the Heat Bank according to the temperature readings
// The Heat Bank has to run at a higher temperature when the CH is on to ensure that the feeds for the radiators and underfloor 
// are hot enough. There is an input connected to the time controller switch so that if that has been used to turn on the CH then
// the Boiler will operate at the right setpoints

// The boiler state and CH state is reported to MQTT regularly and weather and time compensation data is subscribed to.
// Weather and time data is generated by Node-Red, this is used to manage the boiler. There's no point having the 
// boiler run in the middle of the night as the tank cools, or first thing in the morning when the
// sun will shine and heat the tank through the solar thermal and PV diversion. If this info isn't available, i.e. network
// down then basic control is used.


// Compiles in the Arduino IDE for the ESP8266

// GPIO Pin assignement
// GPIO 12  Central Heating On/Off
// GPIO 13  1 wire temp sensors (hard wired to 3 pin header on white board)
// GPIO 14  Boiler On/Off
// GPIO 16  Input pin used to sense if the CH has been turned on using the traditional timer
// All other pins have reserved usage

// normally DEBUG is commented out
#define DEBUG
#ifdef DEBUG
// serial.print statements or any other code only used for debugging
#endif


#include <ESP8266WiFi.h>        // esp WiFi
#include <ESP8266WebServer.h>   // esp webserver
#include <PubSubClient.h>       // esp MQTT
#include <OneWire.h>            // One wire
#include <DallasTemperature.h>  // Dallas sensors
#include <Ticker.h>             // esp interupt timers
#include <EEPROM.h>             // used to save state in case of unplanned restart

// Include the web pages for controlling and configuring the system. The webpages are static
// but include Javascript functions and AJAX calls to ensure the data on the pages is current and gets updated.
#include "CHControllerHTML.h"           // the main webpage in PROGMEM
#include "boilersetpointpage.h"         // boiler setpoint page in PROGMEM
#include "tempsensorpage.h"             // temp sensor assignement page in PROGMEM

//Basic Wifi information
const char* ssid = "Home";
const char* password = "Kor14molen";
const char* wifihostname = "CHnHBControl";
int WiFistatus = WL_IDLE_STATUS;     // the Wifi radio's status

// MQTT and emoncms clients
// Callback function header needs to be delared before the pubsubclient constructor to ensure the client reference
// in the callback is valid - apparently
void mqttcallback(char* topic, byte* payload, unsigned int length);
char* MQTTserver = "192.168.0.20";
WiFiClient mqttClient;
PubSubClient MQTTclient(MQTTserver, 1883, mqttcallback, mqttClient);
String brokerclientName = "CHController";


char* emoncmsserver = "192.168.0.20";
WiFiClient emoncmsClient;
// char apikey[] = "acf89ed963d5e992d960581a290d9c4f";   only needed for https interface we're using the socket interface
int emoncmssocketPort = 20100;
char nodeid[] = "11";
String emoncmspayload;

// Webserver
ESP8266WebServer server(8080);                // Webserver. Using port 8080 as 80 is used by emoncms and we need different ports for the router
bool RefreshMainPage = false;
String msg;

// Flags to manage connection status
boolean MQTTConnected = false;

// DS18B20 sensor data 
#define MAXSENSORS 10                  // maximum number of sensors on the one wire bus
byte DS18B20Address [MAXSENSORS][8];  // 8 bytes per address
float temperatures [MAXSENSORS];      // latest temperature readings
int numTempSensors;                   // number of sensors detected  (0 to n)  
int nextTempSensor;                   // used to work out which sensor to read next - read one every main loop
int errorcode;                        // used to hold last error code
String tempindexerror;                // used when the requested index via web page is greater than # of sensors
// indices of the temperatures in the array used for special purposes
int TankTempforBoilerControlIndex = (4-1);  // Index of the temperature reading from the tank used for boiler control
int CHFeedTempIndex = (7-1);                // Index of the temperature of CH feed for display on web page
int UFHFeedTempIndex = (5-1);               // Index of the temperature of UFH feed for display on web page
OneWire oneWire(13);                  // One wire object on IO pin 13
DallasTemperature tempSensors(&oneWire);

// Timers to handle running tasks at certain points in time
Ticker WiFiWaitforConnectedTimer;             // Timer for the 10 second when WiFiBegin has been issued
int reissueWiFiBeginCounter;                  // counter to count out the minutes between retries of WiFiBegin
boolean WiFiRestarting = false;               // used in the connect to WIFI routine to make sure we don't keep issuing WIFI.begins
Ticker RunningFlashTimer;                     // Timer to flash the running LED every 5s
Ticker TempsTimer;                            // Timer for the main loop to get the temperatures 10s
Ticker NextTempTimer;                         // Timer to attach the read next temp routine to
Ticker PublishResultsTimer;                   // Timer to work out when to publish results to emoncms and MQTT
boolean TimetoPublishResults = false;         // flag set by timer routine to let the main routine when to publish
Ticker CHAddHours;                            // Timer used to manage the add hours functionality for CH

// CH and Boiler control. The Heat Bank set points are different for the times when the CH is on and off.
// Higher set points are used when the CH is on to ensure the tank Heat Bank is hot enough at the take off points
// for the UFH and radiators. Keep the setpoints in variables so they could be set through a web interface
int HeatBankMin = 48;                     // Normal Heat Bank minimum temperature for sending Boiler Demand
int HeatBankMax = 51;                     // Normal Heat Bank maximum temperature for stopping Boiler Demand
int HeatBankCHMin = 60;                   // CH Heat Bank minimum temperature for sending Boiler Demand
int HeatBankCHMax = 66;                   // CH Heat Bank minimum temperature for stopping Boiler Demand
int HBBoilerDemandOnSetPoint = HeatBankMin;       // When temp reaches this point the boiler needs to be fired
int HBBoilerDemandOffSetPoint = HeatBankMax;      // When temp reaches this point the boiler can be turned off
bool Vacation = false;                    // set to true when on vacation through webserver, boiler never fires
bool NightTimeOverride = false;           // used to overide the night time firing of the boiler
bool BoilerBasicControlMode = true;       // used to manage the control mode used for the boiler (basic or weather compensated)
String setpointerror = "";                // used to pass errors back to user when set points are daft


// Set up the GPIO pins and flags for controlling the CH and Boiler   
int CH = 12;                  // CH control on GPIO12
int Boiler = 14;              // Boiler control on GPIO14
int CHManual = 16;            // Input to sense if CH has been manually switched on

bool CHOn = false;            // Shows CH state On or Off - false = Off
bool CHOnManual = false;      // Shows CH state due to manual switch
bool CHAddHoursOn = false;    // True when one of the + hours buttons has been pressed and the time hasn't expired
int AddHours = 0;             // Number of hours to run CH for
int HoursToGo = 0;            // Countdown of the hours remaining
int MinutesToGo = 0;          // Countdown of the minutes within the current hour
bool BoilerOn = false;        // shows the Boiler state On or OFF - false = Off

// EEPROM struct and flag for saving the status
bool SaveStatusRequired = false;    // Flag used to indicate a state or set point has changed and the EEPROM needs refreshing
#define CONFIG_VERSION "V40"
struct Status {
  // This is for detecting if the settings have ever been saved
  char version[4];
  // Saved states for boiler and CH
  bool Vacation; bool CHOn;
  // saved setpoints for boiler control
  int HeatBankMin; int HeatBankMax; int HeatBankCHMin; int HeatBankCHMax;
  // Saved state of any added hours for CH to run
  int AddHours; int HoursToGo; int MinutesToGo;
  // indexes for the key temperatures as per the key # in emoncms   
  int TankTempforBoilerControlIndex; int CHFeedTempIndex; int UFHFeedTempIndex;
} SavedStatus = {
  CONFIG_VERSION,
  Vacation, CHOn,
  HeatBankMin, HeatBankMax, HeatBankCHMin, HeatBankCHMax,
  AddHours, HoursToGo, MinutesToGo,
  TankTempforBoilerControlIndex, CHFeedTempIndex, UFHFeedTempIndex                   
};


// Routine to handle MQTT callback
// Currently empty
void mqttcallback(char* topic, byte* payload, unsigned int length) {
  // handle message arrived
}

/*              //// now all in setup code
// Routine to connect to WiFi - tries to connect and if not connected returns.
void ConnecttoWiFi ()
{  
  // Connect to wifi
  // In case of restart after power failure or if the router has rebooted for some reason
  // we need to restart the process of connecting. Firstly we'll do the WIFI.begin and start a timer
  // to wait for a second to see if we are connected, if after a minute no connection has been
  // made then we'll re-issue the WIFI.begin. The main loop will  call this routine whenever
  // the wifi status is not connected and we're not already restarting. Each time we'll check the status.
  if (!WiFiRestarting) {                // Check to see if we're already restarting
#ifdef DEBUG
  Serial.println();
  Serial.println();
  Serial.print("Connecting to ");
  Serial.println(ssid);
#endif
  // We're not already restarting, so need to kick off the process
  // WiFi.begin starts the process which in the background waits for an IP address
  // WiFi.status checks the status until there is an IP address, i.e. a connection
  WiFi.begin(ssid, password);
  WiFiRestarting = true;                // set the flag to say we are restarting

  // Attach the routine to check if we have a connection, routine runs every 1 seconds and reissues a 
  // Wifi.Begin after a minute of not being connected 
  WiFiWaitforConnectedTimer.attach(1, checkWiFiconnected);
  reissueWiFiBeginCounter = 0;          // reset the counter used to check when to reissue a WiFiBegin   
  }

  // if we've already kicked off a restart then there's nothing else for us to do
}

void checkWiFiconnected () {
  WiFistatus = WiFi.status();
  if (WiFistatus == WL_CONNECTED) {
    // WiFi.status has returned that we are connected, i.e. have been allocated an IP address
#ifdef DEBUG
    Serial.println("");
    Serial.println("WiFi connected");  
    Serial.print("This IP address: ");
    Serial.println(WiFi.localIP());
#endif
    WiFiRestarting = false;
    WiFiWaitforConnectedTimer.detach();    // detach this routine so we stop being called every ten seconds.
  } else {
#ifdef DEBUG
    Serial.print(". ");
//    Serial.println(WiFi.localIP());
    Serial.println (WiFi.status());
    Serial.println (WiFistatus);
#endif
    reissueWiFiBeginCounter++;                // increment the counter to check if a minute has passed
    if (reissueWiFiBeginCounter == 60) {       // 1 minute since the WiFi.Begin so need to reissue it
        WiFi.begin(ssid, password);
        reissueWiFiBeginCounter = 0;          // reset the counter so we count back up to  minute
#ifdef DEBUG
    Serial.println("WiFiBegin reissued");
#endif
    }
  }
}
*/

// Routine to connect to MQTT Server as client
// this is used in setup and then again if a publish fails.
// If connect fails then the MQTTConnected flag is set to false
void ConnecttoMQTT ()
{ 
#ifdef DEBUG
  Serial.print("MQTT: Connecting to ");
  Serial.print(MQTTserver);
  Serial.print(" as ");
  Serial.println(brokerclientName);
#endif
  
  if (MQTTclient.connect((char*) brokerclientName.c_str())) {
#ifdef DEBUG
    Serial.println("Connected to MQTT broker");
#endif
    MQTTConnected = true;
  }
  else {
#ifdef DEBUG
    Serial.println("MQTT connect failed");
#endif
    MQTTConnected = false;
  }
}

/* required if usig the https interface, we're using the socket interface, so will connect and disconnect just before publishing
// Routine to connect to emoncms Server as client
// this is used in setup and then again if a data send fails.
// If connect fails the emoncmsConnected flag is set to false
void Connecttoemoncms ()
{ 
#ifdef DEBUG
  Serial.print("Connecting to ");
  Serial.println(emoncmsserver);
#endif
  
  if (emoncmsClient.connect((char*) emoncmsserver,80)) {
#ifdef DEBUG
    Serial.println("Connected to emoncms");
#endif
    emoncmsConnected = true;
  }
  else {
#ifdef DEBUG
    Serial.println("emoncms connect failed");
#endif
    emoncmsConnected = false;
  }
}
*/

// Routine to get the temperature readings, this is called once every 5 seconds by being attached
// to a timer by the setup code. The routine calls the requestTemperatures and attaches the routine
// to do the actual readings to a 750Ms timer so that each sensor is read in turn. The routine to read the
// actual temperatures detaches itself from the 750s timer once all sensors have been read.
void readtemps()
{
  tempSensors.requestTemperatures();          // Send the command to start conversion
  NextTempTimer.attach(0.75, readnexttemp);   // attach the sensor read routine to the 750ms timer
#ifdef DEBUG
  Serial.println("Current temperature readings");
#endif
}

// Routine to get next temperature reading. Called every 750Ms by attaching to a timer.
void readnexttemp()
{
  // get temperature from next sensor and reset counter to zero when all read
  // routine activated every 750Ms
  // a requestTemperatures has been sent with waitForConversion OFF
  // so get temperature reading for next sensor which should be ready

  float temp=(tempSensors.getTempC(DS18B20Address[nextTempSensor]));        

  // if reading is within range for the sensor convert float to int and save it
  if ((temp<125.0) && (temp>-40.0)) temperatures[nextTempSensor]=(temp); 
     else errorcode = 60 + nextTempSensor +1;     
#ifdef DEBUG
  Serial.print("Sensor number: "); Serial.print(nextTempSensor); 
  Serial.print("  temperature: "); Serial.println(temperatures[nextTempSensor]); 
#endif

  // get ready to read the next sensor at the next interval
  // if we have cycled through all of the sensors, detach the routine from the timer
  // the timer for the bigger loop will reattach this routine to the timer again
  if (nextTempSensor == (numTempSensors -1)) {
    nextTempSensor = 0;
    NextTempTimer.detach ();
  }
    else {
    nextTempSensor++;
    tempSensors.requestTemperatures();         // Send the command to start conversion
  }
}

// 3 routines to manage publishing the temperature sensor readings to MQTT and emoncms
// the first is a routine to set the flag to start the publishing. this is triggered by a timer, this
// timer routine can't do the publishing as it would cause a watchdog event and reboot so the actual publish
// routines need to be called from the main loop once the PublishResults flag is set to true.
// the next 2 are routines to publish to MQTT and emoncms. These are called from the main
// loop when the publish results flag has been set by the timer if WiFi is connected.

void setPublishResultsFlag(){
  TimetoPublishResults = true;
}

void publishtoMQTT() {
    // Publish to MQTT

    if (MQTTclient.connected()){
#ifdef DEBUG
      Serial.print("MQTTClient connected: publishing temperature data: ");
#endif
      int j = 0;
      while (j < numTempSensors) {
        String TempC = String (temperatures[j]);
#ifdef DEBUG
        Serial.print("Sensor: "); Serial.print (j); Serial.print("   Temp: "); Serial.println (TempC); 
#endif
        // Now make the MQTT topic which is the static string plus the sensor nr.
        String Topic = "Temp/" + String (j);
        if (MQTTclient.publish((char*) Topic.c_str(), (char*) TempC.c_str())) {
#ifdef DEBUG
        Serial.println("Publish ok");
        Serial.print("Published: ");
        Serial.print(Topic); Serial.print("/"); Serial.println(TempC);
#endif
        }
        else {
#ifdef DEBUG
          Serial.println("Publish failed");
#endif
        }
      j++;
      }
    }

    else {
#ifdef DEBUG
      Serial.println("MQTT Connection failed");
#endif
      MQTTConnected = false;
      ConnecttoMQTT ();                       // reconnect
    }
}

void publishtoemoncms(){
  // Publish to emoncms
  // first build the payload
  emoncmspayload = String(nodeid);        // initialise the payload with first node id, payload is space delimited
  for (int j = 0; j < numTempSensors; j++) {    // then add in the temperature readings from all available sensors
     emoncmspayload = emoncmspayload + " " + String(temperatures[j],1);
  }
  // for the socket version need to add cr newline
  emoncmspayload = emoncmspayload + + "\r\n";

#ifdef DEBUG
  Serial.print("Connecting to emoncms on: ");
  Serial.println(emoncmsserver);
#endif
  if (emoncmsClient.connect((char*) emoncmsserver, emoncmssocketPort)) { 
  // send the payload, nodeid and Temperatures.
  emoncmsClient.print (emoncmspayload);
#ifdef DEBUG
    Serial.print("emoncmsClient connected: publishing data: ");
    Serial.println (emoncmspayload); 
#endif
    // close connection to emoncms socket
    emoncmsClient.stop();
  }
  else {
    #ifdef DEBUG
    Serial.println("emoncms connect failed");
    #endif
  }
}

// Routines to manage the webserver pages

void handleroot() 
{
  server.send(200, "text/html", CHControllerMainPage);
#ifdef DEBUG
  Serial.println("Webserver root request");
#endif
}

void handleconfig()
{ 
    String message = "\nConfig results\n";
    message += "URI: ";
    message += server.uri();
    message += "\nMethod: ";
    message += ( server.method() == HTTP_GET ) ? "GET" : "POST";
    message += "\nArguments: ";
    message += server.args();
    message += "\n";

    for ( uint8_t i = 0; i < server.args(); i++ ) {
      message += " " + server.argName ( i ) + ": " + server.arg ( i ) + "\n";
    }
#ifdef DEBUG
    Serial.println(message);
#endif
    if (server.argName (0) == "CHSwitch") {
        SaveStatusRequired = true;                  // CH On or CH Off Button pressed therefore need to save state
      if (server.arg (0) == "On") {
        CHOn = true;                                // Set flag for CH On state - this will change the main button to OFF
        Vacation = false;                           // CH turned on so need to turn the Boiler override off
        digitalWrite(CH, HIGH);                     // turn the CH on (HIGH is the voltage level)
          // clear any added hours and detach the timer
          AddHours=0;
          HoursToGo=0;
          MinutesToGo=0;
          CHAddHours.detach ();                     // and detach any timer for additional hours
      }
      if (server.arg (0) == "Off") { 
        CHOn = false;                               // Set flag for CH state - this will change the main button to ON
        digitalWrite(CH, LOW);                      // turn the CH off (LOW is the voltage level)
          // clear any added hours and detach the timer
          AddHours=0;
          HoursToGo=0;
          MinutesToGo=0;
          CHAddHours.detach ();                     // and detach any timer for additional hours
      }
    server.send(200, "text/html", CHControllerMainPage);
    }
    if (server.argName (0) == "AddHours") {
        #ifdef DEBUG
          Serial.print ("One of the add hours buttons clicked: +");
          Serial.println(server.arg (0).toInt());
        #endif
        CHAddHoursOn = true;                        // set flag to indicate add hours is on
        CHOn = true;                                // Set flag for CH On state - this will change the main button to OFF
        Vacation = false;                           // CH turned on so need to turn the Boiler override off
        digitalWrite(CH, HIGH);                     // turn the CH on (HIGH is the voltage level)

        // Save the state in EEPROM in case there is a restart.
        SaveStatusRequired = true;
        
        CHAddHours.detach ();                       // need to detach any previous timer before attaching the new one
  
      // attach the set CHOff routine to timer for one hour and set the number of hours to be on
      AddHours = server.arg (0).toInt();            // Set the number of hours to add - CHDecrementTimeToGo decrements this
      HoursToGo = AddHours-1;
      MinutesToGo = 59;
      CHAddHours.attach(60, CHDecrementTimeToGo);          // Start the timer to count down the hours and minutes. 
      server.send(200, "text/html", CHControllerMainPage);
    }

    if (server.argName (0) == "Override") {
      if (server.arg (0) == "On") {
        if (!CHOn) {
          Vacation = true;                 // CH must be off to set the Boiler override
          // Save the new values in EEPROM in case there is a restart.
          SaveStatusRequired = true;
        }
      }
      if (server.arg (0) == "Off") {
        Vacation = false;
        // Save the new values in EEPROM in case there is a restart.
        SaveStatusRequired = true;
      }
    server.send(200, "text/html", CHControllerMainPage);
    }
    // send the main page

    if (server.argName (0) == "BoilerSetpointConfig") {
      server.send(200, "text/html", BoilerSetpointConfiguration);
    }
    if (server.argName (0) == "TempSensorAssignment") {
      server.send(200, "text/html", TempSensorAssignment);
    }
}
  
void handleboilersetpoints()
{
    if (server.argName (0) == "BoilerSPs") {
      int HBMin = server.arg (1).toInt();
      int HBMax = server.arg (2).toInt();
      int HBCHMin = server.arg (3).toInt();
      int HBCHMax = server.arg (4).toInt();
      if (HBMin >= (HBMax -2) || HBCHMin >= (HBCHMax -2)) { 
        setpointerror = "Minimum set points should be at least 2 lower than maximum";
      } else {
      HeatBankMin = HBMin;
      HeatBankMax = HBMax;
      HeatBankCHMin = HBCHMin;
      HeatBankCHMax = HBCHMax;
      setpointerror = "";

      // Save the new values in EEPROM in case there is a restart. This is done in the main loop.
      SaveStatusRequired = true;
      }
    }
    // send revised page
    server.send(200, "text/html", BoilerSetpointConfiguration);
}

void handletempsensorassignment()
{
    if (server.argName (0) == "TempIndexes") {
      int BoilerIndex = server.arg (1).toInt()-1;
      int CHIndex = server.arg (2).toInt()-1;
      int UFHIndex = server.arg (3).toInt()-1;
      if ((BoilerIndex >= numTempSensors) || (CHIndex >= numTempSensors) || (UFHIndex >= numTempSensors)) {
        tempindexerror = "Index greater than number of sensors. Maximum is: " + String (numTempSensors); 
      } else {
        TankTempforBoilerControlIndex = BoilerIndex;
        CHFeedTempIndex = CHIndex;
        UFHFeedTempIndex = UFHIndex;
        tempindexerror = ""; 
      }

      // Save the new values in EEPROM in case there is a restart.
      SaveStatusRequired = true;
    }
    // send revised page
    server.send(200, "text/html", TempSensorAssignment);
}

void CHDecrementTimeToGo ()                            // goes off every minute ot count down the time
{ 
  MinutesToGo--;                              // decrement the number of minutes in the current hour
  if (MinutesToGo == 0) {                     // check if minutes to go is zero and either stop or decrement hours to go
    if (HoursToGo == 0) {                     
      // if HoursToGo is zero then we're done, turn off CH as the hours to go are also now zero
      CHOn = false;                           // Set flag for CH state - this will change the main button to ON
      digitalWrite(CH, LOW);                  // turn the CH off (LOW is the voltage level)
      // Detach the timer routine (this one)
      CHAddHours.detach ();
      AddHours = 0;
      SaveStatusRequired = true;              // Save the state CH now Off      
    } else {
      HoursToGo--;                            // otherwise knock an hour off the time to go
      MinutesToGo = 59;                       // and set up minutes to go for another hour
    }
  }
}

void GetSavedStatus ()
{
  // This routine reads the EPROM bytes to get the status that was saved
  // Used in setup to make sure that on reboot we don't revert to the defaults
  // Checks to see if any status has been saved - if not routine calls SaveStatus which
  // will save the defaults. This deals with the case where the system has been reflashed
  // EEPROM.begin will be called in the setup code.

  // First check if the status has ever been saved by checking for version number. Status is always 1st in EEPROM
  if (EEPROM.read(0) == CONFIG_VERSION[0] &&
    EEPROM.read(1) == CONFIG_VERSION[1] &&
    EEPROM.read(2) == CONFIG_VERSION[2]) {
    for (unsigned int t=0; t<sizeof(SavedStatus); t++)
      *((char*)&SavedStatus + t) = EEPROM.read(t);
    CHOn = SavedStatus.CHOn;   
    Vacation = SavedStatus.Vacation;
    HeatBankMin = SavedStatus.HeatBankMin;
    HeatBankMax = SavedStatus.HeatBankMax;
    HeatBankCHMin = SavedStatus.HeatBankCHMin;
    HeatBankCHMax = SavedStatus.HeatBankCHMax;
    AddHours = SavedStatus.AddHours; 
    HoursToGo = SavedStatus.HoursToGo; 
    MinutesToGo = SavedStatus.MinutesToGo;
    TankTempforBoilerControlIndex = SavedStatus.TankTempforBoilerControlIndex; 
    CHFeedTempIndex = SavedStatus.CHFeedTempIndex; 
    UFHFeedTempIndex = SavedStatus.UFHFeedTempIndex;

#ifdef DEBUG
    Serial.print ("\nRead status (location:value pairs): ");
    Serial.print (0); Serial.print(":"); Serial.print(Vacation); Serial.print("  ");
    Serial.print (1); Serial.print(":"); Serial.print(CHOn); Serial.println ();
    Serial.print (2); Serial.print(":"); Serial.print(HeatBankMin); Serial.print("  ");
    Serial.print (3); Serial.print(":"); Serial.print(HeatBankMax); Serial.print("  ");
    Serial.print (4); Serial.print(":"); Serial.print(HeatBankCHMin); Serial.print("  ");
    Serial.print (5); Serial.print(":"); Serial.print(HeatBankCHMax); Serial.println("  ");
    Serial.print (6); Serial.print(":"); Serial.print(AddHours); Serial.print("  ");
    Serial.print (7); Serial.print(":"); Serial.print(HoursToGo); Serial.print("  ");
    Serial.print (8); Serial.print(":"); Serial.print(MinutesToGo); Serial.println("  ");
    Serial.print (9); Serial.print(":"); Serial.print(TankTempforBoilerControlIndex); Serial.print("  ");
    Serial.print (10); Serial.print(":"); Serial.print(CHFeedTempIndex); Serial.print("  ");
    Serial.print (11); Serial.print(":"); Serial.print(UFHFeedTempIndex); Serial.println("  ");
#endif  
  } else {
    // Save the status - this will save the defaults
    SaveStatus ();
  }
}

// Routine to save the current status of the system
void SaveStatus (){
  // update the values in the saved status struct
    SavedStatus.CHOn = CHOn;   
    SavedStatus.Vacation = Vacation;
    SavedStatus.HeatBankMin = HeatBankMin;
    SavedStatus.HeatBankMax = HeatBankMax;
    SavedStatus.HeatBankCHMin = HeatBankCHMin;
    SavedStatus.HeatBankCHMax = HeatBankCHMax;
    SavedStatus.AddHours = AddHours; 
    SavedStatus.HoursToGo = HoursToGo; 
    SavedStatus.MinutesToGo = MinutesToGo;
    SavedStatus.TankTempforBoilerControlIndex = TankTempforBoilerControlIndex; 
    SavedStatus.CHFeedTempIndex = CHFeedTempIndex; 
    SavedStatus.UFHFeedTempIndex = UFHFeedTempIndex;
  
  for (unsigned int t=0; t<sizeof(SavedStatus); t++)
    EEPROM.write(t, *((char*)&SavedStatus + t));

  EEPROM.commit();
  
#ifdef DEBUG
    Serial.print ("\nWritten status (location:value pairs): ");
    Serial.print (0); Serial.print(":"); Serial.print(Vacation); Serial.print("  ");
    Serial.print (1); Serial.print(":"); Serial.print(CHOn); Serial.println ();
    Serial.print (2); Serial.print(":"); Serial.print(HeatBankMin); Serial.print("  ");
    Serial.print (3); Serial.print(":"); Serial.print(HeatBankMax); Serial.print("  ");
    Serial.print (4); Serial.print(":"); Serial.print(HeatBankCHMin); Serial.print("  ");
    Serial.print (5); Serial.print(":"); Serial.print(HeatBankCHMax); Serial.println("  ");
    Serial.print (6); Serial.print(":"); Serial.print(AddHours); Serial.print("  ");
    Serial.print (7); Serial.print(":"); Serial.print(HoursToGo); Serial.print("  ");
    Serial.print (8); Serial.print(":"); Serial.print(MinutesToGo); Serial.println("  ");
    Serial.print (9); Serial.print(":"); Serial.print(TankTempforBoilerControlIndex); Serial.print("  ");
    Serial.print (10); Serial.print(":"); Serial.print(CHFeedTempIndex); Serial.print("  ");
    Serial.print (11); Serial.print(":"); Serial.print(UFHFeedTempIndex); Serial.println("  ");
#endif
}

////////////////////////////  SETUP code  //////////////////////////////////////////////////////////

void setup() {

  // initialize CH and Boiler pins as an output and make sure they are both off.
  pinMode(CH, OUTPUT);
  pinMode (Boiler, OUTPUT);
  digitalWrite(CH, LOW);            // make sure CH is OFF
  digitalWrite(Boiler, LOW);        // make sure Boiler is OFF
  pinMode (CHManual, INPUT);        // get ready to check if the manual switch has been used

// Initialise the comms port and wait a bit.
  Serial.begin(115200);
  delay(10000);                                 // wait for 10 seconds - gives a chance to get the serial window open in environment

  // initialise the EEPROM to save the status and retrieve it on a reboot
  EEPROM.begin (512);
#ifdef DEBUG
  Serial.println ("Getting saved status and setting the current state to the last saved state");
#endif
  GetSavedStatus();                             // set Boiler SPs to saved values and state of the CH and vacation flags
  if (SavedStatus.CHOn) {                       // Need to deal with CHOn true state and turn on the CH
    CHOn = true;                                // Set flag for CH On state - this will change the main button to OFF
    Vacation = false;                           // CH turned on so need to turn the Boiler override off
    digitalWrite(CH, HIGH);                     // turn the CH on (HIGH is the voltage level)
 }
 if (SavedStatus.AddHours > 0) {                // Need to deal with there being a number of hours set for CH On                 
    CHOn = true;                                // Set flag for CH On state - this will change the main button to OFF
    Vacation = false;                           // CH turned on so need to turn the Boiler override off
    digitalWrite(CH, HIGH);                     // turn the CH on (HIGH is the voltage level)

    // Need to attach a timer and set the number of hours to run
    CHAddHours.detach ();                       // need to detach any timer before attaching the new one
  
    // attach the set CHOff routine to timer for one hour and set the number of hours to be on
    AddHours = SavedStatus.AddHours;            // Set the number of hours to add - CHDecrementTimeToGo decrements this
    HoursToGo = AddHours-1;
    MinutesToGo = 59;
    CHAddHours.attach(60, CHDecrementTimeToGo);        // Start the first hour - seems we can't call attach with too big a value.
 } 
#ifdef DEBUG
  Serial.println ("We should have it by now");
#endif
 
  // Now connect to WiFi. 
  // Emoncms connection uses websockets so is connected each time results are sent
  // MQTT connection is checked each time it's time to publish results, so will be (re)connected then
  // force wifi mode to STA only (not Access Point) and set hostname
  // esp WiFi reconnects automatically once it is connected. In the case of a power fail this controller will
  // have restarted before the router gets it's act together, although we may find that we connect to any WiFi
  // extender and will then have multiple fails from publishing until the router is up and the emoncms and MQTT
  // server is conected. This setup code waits until there is a connection and then continues. 
  // IT IS IMPORTANT THAT THE BOILER IS OFF (WHICH IT IS) BEFORE THIS CODE STARTS BECAUSE IT WON'T GO
  // ANY FURHTER UNTIL THERE IS A CONNECTION. THE BOILER IS ONLY TURNED ON AND OFF IN THE LOOP CODE.

//  Serial.setDebugOutput(true);

  // All sorts of shit to try to get a connection
  WiFi.persistent(false);
  WiFi.mode(WIFI_OFF); 
//  WiFi.disconnect();
  WiFi.setOutputPower(0);
  
  WiFi.begin(ssid, password);
//  WiFi.mode(WIFI_STA);              
//  WiFi.hostname("CHnHBController");

  Serial.print("Connecting to WiFi");
  while (WiFi.status() != WL_CONNECTED)
  {
    delay(500);
    Serial.print(".");
  }

  // Set Station mode not access point mode and set the hostname
//  WiFi.mode(WIFI_STA);              
//  WiFi.hostname("CHnHBController");

  Serial.println();
  Serial.print("WiFi Connected, IP address: ");
  Serial.println(WiFi.localIP());

/* need to put this on a timer to publish once when wifi and mqtt are up  
// Send the reason for a reset to MQTT. This will be picked up by NodeRed and an email alert sent
  String ResetReason = ESP.getResetReason();
#ifdef DEBUG
  Serial.print("Reset reason "); Serial.println(ResetReason); 
#endif 

    // Publish to MQTT
    if (MQTTclient.connected()){
#ifdef DEBUG
      Serial.print("MQTTClient connected: publishing reset reason");
#endif
      // Now make the MQTT topic.
      if (MQTTclient.publish("ESPReset/CHController/", (char*) ResetReason.c_str())) {
#ifdef DEBUG
        Serial.println("Publish ok");
#endif
      }
      else {
#ifdef DEBUG
        Serial.println("Publish failed");
#endif
      }
    }

    else {              // MQTT not connected
#ifdef DEBUG
      Serial.println("MQTT Connection failed");
#endif
      MQTTConnected = false;
      ConnecttoMQTT ();                       // reconnect
    }

 /////////////////
 */
  
  // set up the Dallas Temp sensors find out how many are connected and get their Ids
  tempSensors.begin();
  numTempSensors=(tempSensors.getDeviceCount()); 
#ifdef DEBUG
  Serial.print("Number of sensors: "); Serial.println(numTempSensors); 
#endif

  if (numTempSensors > 0) {
    // search for one wire devices and put device addresses in array.
    int j=0;
    while ((j < numTempSensors) && (oneWire.search(DS18B20Address[j])))  j++;

    // now turn off WaitforConversion so that any RequestTemperatures returns asap
    // the read next temperatures routine will be attached to a 750Ms timer
    // so the temperature readings will always be ready
    tempSensors.setWaitForConversion(false);         // turn off the wait for conversion
    tempSensors.setResolution (12);                  // set global resolution to 12 bit

    nextTempSensor=0;     // initialise the counter used to select first sensor to read     
    readtemps();          // and run the main routine to read the temperature sensors

    // now attach the main routine to get the sensor readings to a 10s timer
    // so that it runs once every 10 seconds
    TempsTimer.attach (10, readtemps);  
  }

  // attach the publish results flag routine to a 15 second timer to publish results to emoncms and MQTT
  PublishResultsTimer.attach (15, setPublishResultsFlag);

  // now setup the webserver routines to handle requests
  // call to webserver root page

  server.on("/", handleroot);
  server.on("/config", handleconfig);
  server.on("/boilersetpointconfig", handleboilersetpoints);
  server.on("/tempsensorassignment", handletempsensorassignment);

  server.on("/temps", [](){              // 
    String tempsjson = "{";
    tempsjson += "\"TankTempforBoilerControl\":\"" + String(temperatures[TankTempforBoilerControlIndex]) + "\",";
    tempsjson += "\"CHFeedTemp\":\"" + String(temperatures[CHFeedTempIndex]) + "\",";
    tempsjson += "\"UFHFeedTemp\":\"" + String(temperatures[UFHFeedTempIndex]) + "\"";
    tempsjson += "}";
    server.send(200, "text/plain", (char*) tempsjson.c_str());     // send temps to browser on demand
#ifdef DEBUG
    Serial.println("\nWebserver /temps request");
    Serial.print("Sent emoncmspayload: ");
    Serial.println(tempsjson);
#endif    
  });
  server.on("/states", [](){              // send json formatted reply with states
    String statesjson = "{";
    statesjson += "\"BoilerOn\":";
      if (BoilerOn) statesjson += "\"ON\","; else statesjson += "\"OFF\",";
    statesjson += "\"Vacation\":";
      if (Vacation) statesjson += "\"ON\","; else statesjson += "\"OFF\",";
    statesjson += "\"CHOn\":";
      if (CHOn) statesjson += "\"ON\","; else statesjson += "\"OFF\",";
    statesjson += "\"CHOnManual\":";
      if (CHOnManual) statesjson += "\"ON\","; else statesjson += "\"OFF\",";
    statesjson += "\"CHAddHoursOn\":";
      if (CHAddHoursOn) statesjson += "\"ON\","; else statesjson += "\"OFF\",";
    statesjson += "\"AddHours\":";
      statesjson += "\"" + String(AddHours) + "\",";
    statesjson += "\"HoursToGo\":";
      statesjson += "\"" + String(HoursToGo) + "\",";
    statesjson += "\"MinutesToGo\":";
      statesjson += "\"" + String(MinutesToGo) + "\"";
    statesjson += "}";
    server.send(200, "text/plain", (char*) statesjson.c_str());     // send json
#ifdef DEBUG
    Serial.println("\nWebserver /states request");
    Serial.print("Sent states in json: ");
    Serial.println(statesjson);
#endif    
  });
  server.on("/setpointerror", [](){              // send json formatted reply with states
    String errorjson = "{";
    errorjson += "\"SetPointError\":";
      errorjson += "\"" + setpointerror + "\"";
    errorjson += "}";
    server.send(200, "text/plain", (char*) errorjson.c_str());     // send json
#ifdef DEBUG
    Serial.println("\nWebserver /set point error request");
    Serial.print("Sent error in json: ");
    Serial.println(errorjson);
#endif    
  });
  server.on("/tempindexerror", [](){              // send json formatted reply with states
    String errorjson = "{";
    errorjson += "\"TempIndexError\":";
      errorjson += "\"" + tempindexerror + "\"";
    errorjson += "}";
    server.send(200, "text/plain", (char*) errorjson.c_str());     // send json
#ifdef DEBUG
    Serial.println("\nWebserver /temp index error request");
    Serial.print("Sent error in json: ");
    Serial.println(errorjson);
#endif    
  });

  server.on("/getboilersetpoints", [](){              // send json formatted reply with states
    String boilersetpointsjson = "{";
    boilersetpointsjson += "\"NormalOn\":";
      boilersetpointsjson += "\"" + String(HeatBankMin) + "\",";
    boilersetpointsjson += "\"NormalOff\":";
      boilersetpointsjson += "\"" + String(HeatBankMax) + "\",";
    boilersetpointsjson += "\"CHOnOn\":";
      boilersetpointsjson += "\"" + String(HeatBankCHMin) + "\",";
    boilersetpointsjson += "\"CHOnOff\":";
      boilersetpointsjson += "\"" + String(HeatBankCHMax) + "\"";
    boilersetpointsjson += "}";
    server.send(200, "text/plain", (char*) boilersetpointsjson.c_str());     // send json
#ifdef DEBUG
    Serial.println("\nWebserver /boiler setpointrequest request");
    Serial.print("Sent in json: ");
    Serial.println(boilersetpointsjson);
#endif    
  });

  server.on("/gettempassignment", [](){              // send json formatted reply with states
    String tempassignmentjson = "{";
    tempassignmentjson += "\"TankTempforBoilerControlIndex\":";
      tempassignmentjson += "\"" + String(TankTempforBoilerControlIndex+1) + "\",";
    tempassignmentjson += "\"CHFeedTempIndex\":";
      tempassignmentjson += "\"" + String(CHFeedTempIndex+1) + "\",";
    tempassignmentjson += "\"UFHFeedTempIndex\":";
      tempassignmentjson += "\"" + String(UFHFeedTempIndex+1) + "\"";
    tempassignmentjson += "}";
  
    server.send(200, "text/plain", (char*) tempassignmentjson.c_str());     // send json
#ifdef DEBUG
    Serial.println("\nWebserver /temp assignment request");
    Serial.print("Sent in json: ");
    Serial.println(tempassignmentjson);
#endif    
  });

  // start the webserver
  server.begin();
#ifdef DEBUG
  Serial.println("HTTP server started");
#endif

  // Connect to MQTT server on pi hub
  ConnecttoMQTT ();                       // reconnect


delay (15000);                  // delay 15 seconds to make sure we have some temperatures read in
}


////////////////////////////  Main loop code  //////////////////////////////////////////////////////////


void loop() {

  // handle any client requests to the HTTP server according to the server setup using the .on functions above
  server.handleClient();

  // save status if required due to changes to state or Boiler management set points  
  if (SaveStatusRequired) {
    SaveStatus ();
    SaveStatusRequired = false;                   // turn off the flag so we don't keep doing it in the loop function
  }

// Check to see if the manual switch has been used to turn on the CH
// In this case the CH and Boiler pumps will be running already we just need 
// to make sure that the Boiler is in CH mode not in normal mode.
// The input will be on also when the web interface has been used to turn the heating on
// so we only need to set the flag iff the CHOn flag isn't on
  if (digitalRead(CHManual)) {
#ifdef DEBUG
  if (!CHOnManual && !CHOn) Serial.println("Manual CH switch now ON");
#endif
      CHOnManual = true;
      Vacation = false;                           // CH turned on so need to turn the Boiler override off
  } else {
#ifdef DEBUG
  if (CHOnManual) Serial.println("Manual CH switch now OFF");
#endif
      CHOnManual = false;
  }
  
// Make sure the Boiler set points are the right ones - doing it here every time reduces code in the web handler
// Need to use the CH setpoints if the CH has been turned on by the web interface or the Timer on the tank
if (CHOn || CHOnManual) {
    HBBoilerDemandOnSetPoint = HeatBankCHMin;   // Boiler demand on set point is CH Min set point
    HBBoilerDemandOffSetPoint = HeatBankCHMax;  // Boiler demand off set point is CH Max set point
  } else {
    HBBoilerDemandOnSetPoint = HeatBankMin;     // Boiler demand on set point is normal Min set point
    HBBoilerDemandOffSetPoint = HeatBankMax;    // Boiler demand off set point is normal Max set point
}

// Manage the boiler based on the settings for the boiler demand set points
  if (!Vacation) {                              // Only turn boiler on if we're not on vacation
  // Turn the boiler on when the tank top temperature is lower than the on demand low setpoint
    if (!BoilerOn && temperatures[TankTempforBoilerControlIndex] <= HBBoilerDemandOnSetPoint) {
      digitalWrite(Boiler, HIGH);               // turn the Boiler on (HIGH sends the signal to the relay board)
      BoilerOn = true;
#ifdef DEBUG
     Serial.println (temperatures[TankTempforBoilerControlIndex]);
     Serial.println ("Boiler On");
#endif
    }

    // Turn the boiler off when the tank top temperature is higher than the on demand high setpoint
    if (BoilerOn && temperatures[TankTempforBoilerControlIndex] >= HBBoilerDemandOffSetPoint) {
     digitalWrite(Boiler, LOW);               // turn the Boiler off (LOW sends the signal to the relay board)
     BoilerOn = false;
#ifdef DEBUG
     Serial.println (temperatures[TankTempforBoilerControlIndex]);
     Serial.println ("Boiler Off");
#endif
    }
  } 

// Main results publishing section. When the flag is set by the timer routine this section 
// is executed. Whatever happens the flag is set to false at the end.

  if (TimetoPublishResults == true) {
#ifdef DEBUG
    Serial.println ("Publishing data");
#endif
    publishtoMQTT();                        // publish to MQTT
    publishtoemoncms();                     // publish to emoncms

    // reset the flag which will be set again once the timer triggers
    TimetoPublishResults = false;
  }

  // let anything else in the esp and libraries have a chance to execute
  yield();
}





